"""I/O connectors of mysql."""

from typing import Dict, Union

import apache_beam as beam
from apache_beam.io import iobase
from apache_beam.options.value_provider import ValueProvider
from apache_beam.pvalue import PCollection
from apache_beam.transforms.core import PTransform

from beam_mysql.connector import splitters
from beam_mysql.connector.client import MySQLClient
from beam_mysql.connector.source import MySQLSource
from beam_mysql.connector.utils import get_runtime_value


class ReadFromMySQL(PTransform):
    """Create PCollection from MySQL."""

    def __init__(
        self,
        query: Union[str, ValueProvider],
        host: Union[str, ValueProvider],
        database: Union[str, ValueProvider],
        user: Union[str, ValueProvider],
        password: Union[str, ValueProvider],
        port: Union[int, ValueProvider] = 3306,
        splitter=splitters.NoSplitter(),
    ):
        super().__init__()
        self._query = query
        self._host = host
        self._database = database
        self._user = user
        self._password = password
        self._port = port
        self._splitter = splitter

    def expand(self, pcoll: PCollection) -> PCollection:
        return pcoll | iobase.Read(
            MySQLSource(
                self._query,
                self._host,
                self._database,
                self._user,
                self._password,
                self._port,
                self._splitter,
            )
        )


class WriteToMySQL(PTransform):
    """Write dict rows to MySQL."""

    def __init__(
        self,
        host: Union[str, ValueProvider],
        database: Union[str, ValueProvider],
        table: Union[str, ValueProvider],
        user: Union[str, ValueProvider],
        password: Union[str, ValueProvider],
        port: Union[int, ValueProvider] = 3306,
        batch_size: int = 1000,
    ):
        super().__init__()
        self._host = host
        self._database = database
        self._table = table
        self._user = user
        self._password = password
        self._port = port
        self._batch_size = batch_size

    def expand(self, pcoll: PCollection) -> PCollection:
        return pcoll | beam.ParDo(
            _WriteToMySQLFn(
                self._host,
                self._database,
                self._table,
                self._user,
                self._password,
                self._port,
                self._batch_size,
            )
        )


class _WriteToMySQLFn(beam.DoFn):
    """DoFn for WriteToMySQL."""

    def __init__(
        self,
        host: Union[str, ValueProvider],
        database: Union[str, ValueProvider],
        table: Union[str, ValueProvider],
        user: Union[str, ValueProvider],
        password: Union[str, ValueProvider],
        port: Union[int, ValueProvider],
        batch_size: int,
    ):
        super().__init__()
        self._host = host
        self._database = database
        self._table = table
        self._user = user
        self._password = password
        self._port = port
        self._batch_size = batch_size

        self._config = {
            "host": self._host,
            "database": self._database,
            "user": self._user,
            "password": self._password,
            "port": self._port,
        }

    def start_bundle(self):
        self._build_value()
        self._columns_and_values = dict()

    def process(self, element: Dict, *args, **kwargs):
        columns = []
        values = []
        for column, value in element.items():
            columns.append(column)
            values.append(value)

        column_str = ", ".join(columns)
        value_str = ", ".join(
            [
                f"{'NULL' if value is None else value}"
                if isinstance(value, (type(None), int, float))
                else f"'{value}'"
                for value in values
            ]
        )

        if column_str not in self._columns_and_values:
            self._columns_and_values[column_str] = []

        self._columns_and_values[column_str].append(f"({value_str})")

        if len(self._columns_and_values) > self._batch_size:
            for column_str in self._columns_and_values.keys():
                if len(self._columns_and_values[column_str]) > 0:
                    self._client.record_loader(self._build_query(column_str, self._columns_and_values[column_str]))
                    self._columns_and_values[column_str].clear()

    def finish_bundle(self):
        for column_str in self._columns_and_values.keys():
            if len(self._columns_and_values[column_str]) > 0:
                self._client.record_loader(self._build_query(column_str, self._columns_and_values[column_str]))
                self._columns_and_values[column_str].clear()

    def _build_query(self, column_str, values_str):
        return f"INSERT INTO {self._config['database']}.{self._table}({column_str}) VALUES {','.join(values_str)};"

    def _build_value(self):
        for k, v in self._config.items():
            self._config[k] = get_runtime_value(v)
        self._table = get_runtime_value(self._table)
        self._batch_size = get_runtime_value(self._batch_size)

        self._client = MySQLClient(self._config)



"""A client of mysql."""

import re
from logging import INFO, getLogger
from typing import Dict, Generator

import mysql.connector
from mysql.connector.errors import Error as MySQLConnectorError

from beam_mysql.connector.errors import MySQLClientError

_SELECT_STATEMENT = "SELECT"
_INSERT_STATEMENT = "INSERT"

logger = getLogger(__name__)
logger.setLevel(INFO)


class MySQLClient:
    """A mysql client object."""

    def __init__(self, config: Dict):
        self._config = config
        self._validate_config(self._config)

    def record_generator(
        self, query: str, dictionary=True
    ) -> Generator[Dict, None, None]:
        """
        Generate dict record from raw data on mysql.

        Args:
            query: query with select statement
            dictionary: the type of result is dict if true else tuple

        Returns:
            dict record

        Raises:
            ~beam_mysql.connector.errors.MySQLClientError
        """
        self._validate_query(query, _SELECT_STATEMENT)

        with _MySQLConnection(self._config) as conn:
            # buffered is false because it can be assumed that the data size is too large
            cur = conn.cursor(buffered=False, dictionary=dictionary)

            try:
                cur.execute(query)
                logger.info(f"Successfully execute query: {query}")

                for record in cur:
                    yield record
            except MySQLConnectorError as e:
                raise MySQLClientError(
                    f"Failed to execute query: {query}, Raise exception: {e}"
                )

            cur.close()

    def counts_estimator(self, query: str) -> int:
        """
        Make a estimate of the total number of records.

        Args:
            query: query with select statement

        Returns:
            the total number of records

        Raises:
            ~beam_mysql.connector.errors.MySQLClientError
        """
        self._validate_query(query, _SELECT_STATEMENT)
        count_query = f"SELECT COUNT(*) AS count FROM ({query}) as subq"

        with _MySQLConnection(self._config) as conn:
            # buffered is false because it can be assumed that the data size is too large
            cur = conn.cursor(buffered=False, dictionary=True)

            try:
                cur.execute(count_query)
                logger.info(f"Successfully execute query: {count_query}")

                record = cur.fetchone()
            except MySQLConnectorError as e:
                raise MySQLClientError(
                    f"Failed to execute query: {count_query}, Raise exception: {e}"
                )

            cur.close()

            return record["count"]

    def rough_counts_estimator(self, query: str) -> int:
        """
        Make a rough estimate of the total number of records.
        To avoid waiting time by select counts query when the data size is too large.

        Args:
            query: query with select statement

        Returns:
            the total number of records

        Raises:
            ~beam_mysql.connector.errors.MySQLClientError
        """
        self._validate_query(query, _SELECT_STATEMENT)
        count_query = f"EXPLAIN SELECT * FROM ({query}) as subq"

        with _MySQLConnection(self._config) as conn:
            # buffered is false because it can be assumed that the data size is too large
            cur = conn.cursor(buffered=False, dictionary=True)

            try:
                cur.execute(count_query)
                logger.info(f"Successfully execute query: {count_query}")

                records = cur.fetchall()

                total_number = 0

                for record in records:
                    # Query of the argument should be "DERIVED" because it is sub query of explain select.
                    # Count query should be "PRIMARY" or "SIMPLE" because it is not sub query.
                    if record["select_type"] in ("PRIMARY", "SIMPLE"):
                        total_number = int(record["rows"])
            except MySQLConnectorError as e:
                raise MySQLClientError(
                    f"Failed to execute query: {count_query}, Raise exception: {e}"
                )

            cur.close()

            if total_number <= 0:
                raise mysql.connector.errors.Error(
                    f"Failed to estimate total number of records. Query: {count_query}"
                )
            else:
                return total_number

    def record_loader(self, query: str):
        """
        Load dict record into mysql.

        Args:
            query: query with insert or update statement

        Raises:
            ~beam_mysql.connector.errors.MySQLClientError
        """
        self._validate_query(query, _INSERT_STATEMENT)

        with _MySQLConnection(self._config) as conn:
            cur = conn.cursor()

            try:
                cur.execute(query)
                conn.commit()
                logger.info(f"Successfully execute query: {query}")
            except MySQLConnectorError as e:
                conn.rollback()
                raise MySQLClientError(
                    f"Failed to execute query: {query}, Raise exception: {e}"
                )

            cur.close()

    @staticmethod
    def _validate_config(config: Dict):
        required_keys = {"host", "port", "database", "user", "password"}
        if not config.keys() == required_keys:
            raise MySQLClientError(
                f"Config is not satisfied. required: {required_keys}, actual: {config.keys()}"
            )

    @staticmethod
    def _validate_query(query: str, statement: str) -> None:
        def _remove_comments_and_cte(query):
            # delete comments
            query = re.sub(r"--.*\n", " ", query)
            query = re.sub(r"/\*[.\s].*\*/", " ", query, flags=re.DOTALL)
            # delete new line
            query = query.replace("\n", " ")
            # delete cte clause
            query = re.sub(
                r"WITH\s+(?:\w+\s+AS\s+\(.+?\)\s*,?\s*)+",
                "",
                query,
                flags=re.IGNORECASE,
            )
            return query.strip()

        cleansed_query = _remove_comments_and_cte(query)

        if statement and not cleansed_query.lower().startswith(statement.lower()):
            raise MySQLClientError(
                f"Query expected to start with {statement} statement. Query: {query}"
            )


class _MySQLConnection:
    """A wrapper object to connect mysql."""

    def __init__(self, _config: Dict):
        self._config = _config

    def __enter__(self):
        try:
            self.conn = mysql.connector.connect(**self._config)
            return self.conn
        except MySQLConnectorError as e:
            raise MySQLClientError(f"Failed to connect mysql, Raise exception: {e}")

    def __exit__(self, exception_type, exception_value, traceback):
        self.conn.close()



"""Beam - MySQL Connector for setup file."""

import os
import subprocess

from setuptools import Command, find_packages, setup

SRC_DIR_NAME = "beam_mysql"
EXAMPLES_DIR_NAME = "examples"
TESTS_DIR_NAME = "tests"


class SimpleCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


class LintCommand(SimpleCommand):
    def run(self):
        subprocess.run(
            [
                "black",
                SRC_DIR_NAME,
                EXAMPLES_DIR_NAME,
                TESTS_DIR_NAME,
                "-l",
                "120",
                "--check",
                "--diff",
            ]
        )
        subprocess.run(["mypy", SRC_DIR_NAME, EXAMPLES_DIR_NAME, TESTS_DIR_NAME])


class FormatCommand(SimpleCommand):
    def run(self):
        subprocess.run(
            ["black", SRC_DIR_NAME, EXAMPLES_DIR_NAME, TESTS_DIR_NAME, "-l", "120"]
        )


def get_version():
    global_names = {}
    exec(
        open(
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "beam_mysql/__init__.py"
            )
        ).read(),
        global_names,
    )
    return global_names["__version__"]


with open("README.md") as f:
    README = f.read()


PACKAGE_NAME = "beam-mysql-connector"
PACKAGE_VERSION = get_version()
PACKAGE_DESCRIPTION = "Apache Beam I/O connector designed for accessing MySQL databases."
PACKAGE_URL = "https://github.com/esakik/beam-mysql-connector"
PACKAGE_DOWNLOAD_URL = "https://pypi.python.org/pypi/beam-mysql-connector"
PACKAGE_AUTHOR = "esakik"
PACKAGE_EMAIL = "esakik.dev@gmail.com"
PACKAGE_KEYWORDS = "apache beam mysql connector"
PACKAGE_LONG_DESCRIPTION = README

REQUIRED_PACKAGES = [
    "apache-beam>=2.25.0",
    "mysql-connector-python>=8.1.0",
]

setup(
    name=PACKAGE_NAME,
    version=PACKAGE_VERSION,
    description=PACKAGE_DESCRIPTION,
    long_description=PACKAGE_LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url=PACKAGE_URL,
    download_url=PACKAGE_DOWNLOAD_URL,
    author=PACKAGE_AUTHOR,
    author_email=PACKAGE_EMAIL,
    packages=find_packages(),
    install_requires=REQUIRED_PACKAGES,
    python_requires=">=3.8,<3.12",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    license="MIT",
    keywords=PACKAGE_KEYWORDS,
    cmdclass=dict(lint=LintCommand, fmt=FormatCommand),
)
